@@ -22,19 +22,23 @@ import java.security.AccessController;
 import java.security.PrivilegedActionException;
 import java.security.PrivilegedExceptionAction;
 import java.util.*;
+import org.apache.felix.framework.capabilityset.Attribute;
+import org.apache.felix.framework.capabilityset.Capability;
+import org.apache.felix.framework.capabilityset.Directive;
+import org.apache.felix.framework.resolver.Module;
+import org.apache.felix.framework.resolver.Wire;
 
 import org.apache.felix.framework.util.MapToDictionary;
 import org.apache.felix.framework.util.StringMap;
 import org.apache.felix.framework.util.Util;
-import org.apache.felix.moduleloader.IModule;
-import org.apache.felix.moduleloader.IWire;
 import org.osgi.framework.*;
+import org.osgi.framework.BundleReference;
 
 class ServiceRegistrationImpl implements ServiceRegistration
 {
     // Service registry.
     private final ServiceRegistry m_registry;
-    // Bundle implementing the service.
+    // Bundle providing the service.
     private final Bundle m_bundle;
     // Interfaces associated with the service object.
     private final String[] m_classes;
@@ -68,9 +72,6 @@ class ServiceRegistrationImpl implements ServiceRegistration
 
         // This reference is the "standard" reference for this
         // service and will always be returned by getReference().
-        // Since all reference to this service are supposed to
-        // be equal, we use the hashcode of this reference for
-        // a references to this service in ServiceReference.
         m_ref = new ServiceReferenceImpl();
     }
 
@@ -146,19 +147,32 @@ class ServiceRegistrationImpl implements ServiceRegistration
     **/
     private boolean isClassAccessible(Class clazz)
     {
-        try
+        // We need to see if the class loader of the service object has
+        // access to the specified class; however, we may not have a service
+        // object. If we only have service factory, then we will assume two
+        // different scenarios:
+        // 1. The service factory is provided by the bundle providing the
+        //    service.
+        // 2. The service factory is NOT provided by the bundle providing
+        //    the service.
+        // For case 1, we will use the class loaded of the service factory
+        // to find the class. For case 2, we will assume we have an extender
+        // at work here and always return true, since we have no real way of
+        // knowing the wiring of the provider unless we actually get the
+        // service object, which defeats the lazy aspect of service factories.
+
+        // Case 2.
+        if ((m_factory != null)
+            && (m_factory.getClass().getClassLoader() instanceof BundleReference)
+            && !((BundleReference) m_factory.getClass()
+                .getClassLoader()).getBundle().equals(m_bundle))
         {
-            // Try to load from the service object or service factory class.
-            Class sourceClass = (m_factory != null)
-                ? m_factory.getClass() : m_svcObj.getClass();
-            Class targetClass = Util.loadClassUsingClass(sourceClass, clazz.getName());
-            return (targetClass == clazz);
+            return true;
         }
-        catch (Exception ex)
-        {
-            // Ignore this and return false.
-        }
-        return false;
+
+        // Case 1.
+        Class sourceClass = (m_factory != null) ? m_factory.getClass() : m_svcObj.getClass();
+        return Util.loadClassUsingClass(sourceClass, clazz.getName(), Felix.m_secureAction) == clazz;
     }
 
     Object getProperty(String key)
@@ -302,7 +316,8 @@ class ServiceRegistrationImpl implements ServiceRegistration
         {
             for (int i = 0; i < m_classes.length; i++)
             {
-                Class clazz = Util.loadClassUsingClass(svcObj.getClass(), m_classes[i]);
+                Class clazz = Util.loadClassUsingClass(
+                    svcObj.getClass(), m_classes[i], Felix.m_secureAction);
                 if ((clazz == null) || !clazz.isAssignableFrom(svcObj.getClass()))
                 {
                     if (clazz == null)
@@ -363,7 +378,11 @@ class ServiceRegistrationImpl implements ServiceRegistration
         }
     }
 
-    class ServiceReferenceImpl implements ServiceReference
+    //
+    // ServiceReference implementation
+    //
+
+    class ServiceReferenceImpl implements ServiceReference, Capability
     {
         private ServiceReferenceImpl() {}
 
@@ -372,6 +391,46 @@ class ServiceRegistrationImpl implements ServiceRegistration
             return ServiceRegistrationImpl.this;
         }
 
+        //
+        // Capability methods.
+        //
+
+        public Module getModule()
+        {
+            throw new UnsupportedOperationException("Not supported yet.");
+        }
+
+        public String getNamespace()
+        {
+            return "service-reference";
+        }
+
+        public Directive getDirective(String name)
+        {
+            return null;
+        }
+
+        public List<Directive> getDirectives()
+        {
+            return Util.m_emptyList;
+        }
+
+        public Attribute getAttribute(String name)
+        {
+            Object value = ServiceRegistrationImpl.this.getProperty(name);
+            return (value == null) ? null : new Attribute(name, value, false);
+        }
+
+        public List<Attribute> getAttributes()
+        {
+            return Util.m_emptyList;
+        }
+
+        public List<String> getUses()
+        {
+            return Util.m_emptyList;
+        }
+
         public Object getProperty(String s)
         {
             return ServiceRegistrationImpl.this.getProperty(s);
@@ -421,12 +480,12 @@ class ServiceRegistrationImpl implements ServiceRegistration
             // Get the package.
             String pkgName =
                 Util.getClassPackage(className);
-            IModule requesterModule = ((BundleImpl) requester).getCurrentModule();
+            Module requesterModule = ((BundleImpl) requester).getCurrentModule();
             // Get package wiring from service requester.
-            IWire requesterWire = Util.getWire(requesterModule, pkgName);
+            Wire requesterWire = Util.getWire(requesterModule, pkgName);
             // Get package wiring from service provider.
-            IModule providerModule = ((BundleImpl) m_bundle).getCurrentModule();
-            IWire providerWire = Util.getWire(providerModule, pkgName);
+            Module providerModule = ((BundleImpl) m_bundle).getCurrentModule();
+            Wire providerWire = Util.getWire(providerModule, pkgName);
 
             // There are four situations that may occur here:
             //   1. Neither the requester, nor provider have wires for the package.
